@@ -188,6 +188,7 @@ int do_move_pid_main(const char *controller, const char *cgroup, struct ucred p,
 	nih_local char *c = NULL;
 	char *tok;
 	int ret;
+	int while_ret = 0;
 
 	if (!sane_cgroup(cgroup)) {
 		nih_error("%s: unsafe cgroup", __func__);
@@ -214,15 +215,16 @@ int do_move_pid_main(const char *controller, const char *cgroup, struct ucred p,
 	tok = strtok(c, ",");
 	while (tok) {
 		ret = per_ctrl_move_pid_main(tok, cgroup, p, r, v, escape);
-		if (ret == -2)  // permission denied - ignore for group requests
-			goto next;
-		if (ret != 0)
-			return -1;
-next:
+
+		/* Save error for later (but ignore permission denied, -2),
+		   but try to complete rest of moves anyway */
+		if (ret != 0 && ret != -2)
+			while_ret = -1;
+
 		tok = strtok(NULL, ",");
 	}
 
-	return 0;
+	return while_ret;
 }
 
 int move_pid_main(const char *controller, const char *cgroup, struct ucred p,
@@ -523,8 +525,7 @@ int chmod_main(const char *controller, const char *cgroup, const char *file,
 	}
 
 	if (file && ( strchr(file, '/') || strchr(file, '\\')) ) {
-		nih_dbus_error_raise_printf (DBUS_ERROR_INVALID_ARGS,
-				"invalid file");
+		nih_error("%s: invalid file", __func__);
 		return -1;
 	}
 
@@ -575,6 +576,12 @@ int get_value_main(void *parent, const char *controller, const char *cgroup,
 		return -1;
 	}
 
+	if (!path_is_under_taskcg(p.pid, controller, path)) {
+		nih_error("%s: target cgroup is not below r (%d)'s", __func__,
+			r.pid);
+		return -1;
+	}
+
 	/* append the filename */
 	if (strlen(path) + strlen(key) + 2 > MAXPATHLEN) {
 		nih_error("%s: filename too long for cgroup %s key %s", __func__, path, key);
@@ -619,6 +626,12 @@ int set_value_main(const char *controller, const char *cgroup,
 		return -1;
 	}
 
+	if (!path_is_under_taskcg(p.pid, controller, path)) {
+		nih_error("%s: target cgroup is not below r (%d)'s", __func__,
+			r.pid);
+		return -1;
+	}
+
 	/* Check access rights to the cgroup directory */
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 		nih_error("%s: Pid %d may not access %s\n", __func__, r.pid, path);
@@ -836,6 +849,12 @@ int get_tasks_main(void *parent, const char *controller, const char *cgroup,
 		return -1;
 	}
 
+	if (!path_is_under_taskcg(p.pid, controller, path)) {
+		nih_error("%s: target cgroup is not below r (%d)'s", __func__,
+			r.pid);
+		return -1;
+	}
+
 	/* Check access rights to the cgroup directory */
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 		nih_error("%s: Pid %d may not access %s\n", __func__, r.pid, path);
@@ -853,7 +872,8 @@ int get_tasks_main(void *parent, const char *controller, const char *cgroup,
 
 	*pids = NULL;
 	if (file_read_pids(parent, path, pids, &alloced_pids, &nrpids) < 0) {
-		nih_free(*pids);
+		if (*pids)
+			nih_free(*pids);
 		return -1;
 	}
 	return nrpids;
@@ -918,6 +938,12 @@ int collect_tasks(void *parent, const char *controller, const char *cgroup,
 		return -2;
 	}
 
+	if (!path_is_under_taskcg(p.pid, controller, path)) {
+		nih_error("%s: target cgroup is not below r (%d)'s", __func__,
+			r.pid);
+		return -1;
+	}
+
 	/* Check access rights to the cgroup directory */
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 		nih_error("%s: Pid %d may not access %s\n", __func__, r.pid, path);
@@ -946,7 +972,8 @@ int get_tasks_recursive_main(void *parent, const char *controller,
 	if (strcmp(controller, "all") != 0 && !strchr(controller, ',')) {
 		if (collect_tasks(parent, controller, cgroup, p, r, pids,
 				&alloced_pids, &nrpids) < 0) {
-			nih_free(*pids);
+			if (*pids)
+				nih_free(*pids);
 			return -1;
 		}
 		return nrpids;
@@ -967,7 +994,8 @@ int get_tasks_recursive_main(void *parent, const char *controller,
 		if (ret == -2)  // permission denied - ignore
 			goto next;
 		if (ret != 0) {
-			nih_free(*pids);
+			if (*pids)
+				nih_free(*pids);
 			return -1;
 		}
 next:
@@ -994,6 +1022,12 @@ int list_children_main(void *parent, const char *controller, const char *cgroup,
 		return -1;
 	}
 
+	if (!path_is_under_taskcg(p.pid, controller, path)) {
+		nih_error("%s: target cgroup is not below r (%d)'s", __func__,
+			r.pid);
+		return -1;
+	}
+
 	/* Check access rights to the cgroup directory */
 	if (!may_access(r.pid, r.uid, r.gid, path, O_RDONLY)) {
 		nih_error("%s: Pid %d may not access %s\n", __func__, r.pid, path);
@@ -1262,9 +1296,18 @@ int list_keys_main(void *parent, const char *controller, const char *cgroup,
 }
 
 char *extra_cgroup_mounts;
+char *skip_cgroup_mounts;
 
 static int
-my_setter (NihOption *option, const char *arg)
+skip_mounts_set (NihOption *option, const char *arg)
+{
+	skip_cgroup_mounts = NIH_MUST( strdup(arg) );
+
+	return 0;
+}
+
+static int
+extra_mounts_set (NihOption *option, const char *arg)
 {
 	extra_cgroup_mounts = NIH_MUST( strdup(arg) );
 
@@ -1279,8 +1322,10 @@ my_setter (NihOption *option, const char *arg)
 static NihOption options[] = {
 	{ 0, "max-depth", N_("Maximum cgroup depth"),
 		NULL, NULL, &maxdepth, NULL },
+	{ 'M', "skip", N_("Subsystems to not mount"),
+		NULL, "subsystems to mount", NULL, skip_mounts_set },
 	{ 'm', "mount", N_("Extra subsystems to mount"),
-		NULL, "subsystems to mount", NULL, my_setter },
+		NULL, "subsystems to mount", NULL, extra_mounts_set },
 	{ 0, "daemon", N_("Detach and run in the background"),
 		NULL, NULL, &daemonise, NULL },
 	{ 0, "sigstop", N_("Raise SIGSTOP when ready"),
@@ -1326,11 +1371,11 @@ static bool dir_is_sysfs(const char *path)
 
 static void mount_tmpfs_at(const char *path)
 {
-	if (mount("cgroup", path, "tmpfs", 0, "size=10000") < 0) {
+	if (mount("cgroup", path, "tmpfs", 0, "size=10000,mode=0755") < 0) {
 		nih_error("Failed to mount tmpfs at %s: %m", path);
 		exit(1);
 	}
-	nih_debug("Mounted tmpfs onto %s", CGDIR);
+	nih_debug("Mounted tmpfs onto %s", path);
 }
 
 static bool is_ro_mount(const char *path)
@@ -1419,7 +1464,7 @@ main (int argc, char *argv[])
 		return -1;
 	}
 
-	if (collect_subsystems(extra_cgroup_mounts) < 0)
+	if (collect_subsystems(extra_cgroup_mounts, skip_cgroup_mounts) < 0)
 	{
 		nih_fatal("failed to collect cgroup subsystems");
 		exit(1);
