@@ -109,6 +109,14 @@ enum req_type {
 	REQ_TYPE_GET_PID_ABS,
 	REQ_TYPE_PRUNE,
 	REQ_TYPE_LISTCONTROLLERS,
+	REQ_TYPE_LISTKEYS,
+};
+
+struct keys_return_type {
+	char *name;
+	uint32_t uid;
+	uint32_t gid;
+	uint32_t perms;
 };
 
 int get_pid_cgroup_main(void *parent, const char *controller,
@@ -163,6 +171,10 @@ void prune_scm_complete(struct scm_sock_data *data);
 
 int list_controllers_main (void *parent, char ***output);
 
+int list_keys_main (void *parent, const char *controller, const char *cgroup,
+			struct ucred p, struct ucred r,
+			struct keys_return_type ***output);
+void list_keys_scm_complete (struct scm_sock_data *data);
 
 int cgmanager_ping (void *data, NihDBusMessage *message, int junk);
 
@@ -171,6 +183,6 @@ void client_disconnect (DBusConnection *conn);
 
 bool sane_cgroup(const char *cgroup);
 
-#define API_VERSION 8
+#define API_VERSION 10
 
 #endif
