@@ -56,6 +56,21 @@
 #define AGENT SBINDIR "/cgm-release-agent"
 #define AGENT_LINK_PATH "/run/cgmanager/agents"
 
+/* Define pivot_root() if missing from the C library */
+#ifndef HAVE_PIVOT_ROOT
+static int pivot_root(const char * new_root, const char * put_old)
+{
+#ifdef __NR_pivot_root
+return syscall(__NR_pivot_root, new_root, put_old);
+#else
+errno = ENOSYS;
+return -1;
+#endif
+}
+#else
+extern int pivot_root(const char * new_root, const char * put_old);
+#endif
+
 char *all_controllers;
 
 struct controller_mounts {
@@ -67,6 +82,7 @@ struct controller_mounts {
 	struct controller_mounts *comounted;
 	bool premounted;
 	bool visited;
+	bool skip;
 };
 
 static struct controller_mounts *all_mounts;
@@ -124,6 +140,10 @@ bool setup_base_run_path(void)
 		nih_fatal("%s: failed to create /run/cgmanager/fs", __func__);
 		return false;
 	}
+	if (mount("cgmfs", "/run/cgmanager/fs", "tmpfs", 0, "size=100000,mode=0755") < 0) {
+		nih_fatal("%s: failed to mount tmpfs onto /run/cgmanager/fs", __func__);
+		return false;
+	}
 	if (mkdir(AGENT_LINK_PATH, 0755) < 0 && errno != EEXIST) {
 		nih_fatal("%s: failed to create %s", __func__, AGENT_LINK_PATH);
 		return false;
@@ -337,6 +357,11 @@ static bool do_mount_subsys(int i)
 	controller = m->controller;
 	src = m->src;
 
+	if (m->skip) {
+		nih_info("Skipping mount of %s as requested\n", src);
+		return true;
+	}
+
 	if (mkdir(dest, 0755) < 0 && errno != EEXIST) {
 		nih_fatal("Failed to create %s: %s", dest, strerror(errno));
 		return false;
@@ -554,6 +579,15 @@ static bool collect_kernel_subsystems(void)
 			continue;
 		*p = '\0';
 
+		// TODO: How stable is /proc/cgroups interface?
+		// Check the 'enabled' column
+		p = strrchr(p+1, '\t');
+		if (!p)
+			continue;
+
+		if (*(p+1) != '1')
+			continue;
+
 		if (!save_mount_subsys(line)) {
 			nih_fatal("Error storing subsystem %s", line);
 			goto out;
@@ -617,7 +651,7 @@ static void build_controller_mntlist(void)
 	for (i = 0; i < num_controllers; i++) {
 		char *srclist;
 		m = &all_mounts[i];
-		if (m->visited)
+		if (m->visited || m->skip)
 			continue;
 		controller_mnts = realloc(controller_mnts,
 				(num_controller_mnts+1)*(sizeof(char *)));
@@ -658,6 +692,7 @@ static void print_debug_controller_info(void)
 		nih_debug("    src %s path %s options %s",
 			m->src, m->path ? m->path : "(none)", m->options ? m->options : "(none)");
 		nih_debug("    agent: %s", m->agent ? m->agent : "(none)");
+		nih_debug("    skipped: %s", m->skip ? "yes" : "no");
 		nih_debug("    premounted: %s comounted: %s",
 			m->premounted ? "yes" : "no",
 			m->comounted ? m->comounted->controller : "(none)");
@@ -710,7 +745,34 @@ void do_prune_comounts(char *controllers)
 	}
 }
 
-static void build_all_controllers(void)
+/*
+ * @list is a comma-separated list of words.
+ * Return true if @word is in @list.
+ */
+static bool word_in_list(char *word, char *list)
+{
+	char *p = list;
+	size_t wlen;
+
+	if (!list)
+		return false;
+
+	wlen = strlen(word);
+	while (p && *p) {
+		size_t len;
+		char *pe = strchr(p, ',');
+		len = pe ? pe - p : strlen(p);
+		if (len == wlen && strncmp(word, p, len) == 0)
+			return true;
+		if (pe)
+			pe++;
+		p = pe;
+	}
+
+	return false;
+}
+
+static void build_all_controllers(char *skip_mounts)
 {
 	int i;
 	char *c;
@@ -719,6 +781,10 @@ static void build_all_controllers(void)
 		c = all_mounts[i].src;
 		if (strncmp(c, "none,", 5) == 0)
 			c += 5;
+		if (word_in_list(c, skip_mounts)) {
+			all_mounts[i].skip = true;
+			continue;
+		}
 		if (!all_controllers)
 			all_controllers = NIH_MUST( nih_strdup(NULL, c) );
 		else
@@ -728,7 +794,7 @@ static void build_all_controllers(void)
 	do_prune_comounts(all_controllers);
 }
 
-int collect_subsystems(char *extra_mounts)
+int collect_subsystems(char *extra_mounts, char *skip_mounts)
 {
 	/* first collect all already-mounted subsystems */
 	if (!collect_premounted_subsystems())
@@ -751,7 +817,7 @@ int collect_subsystems(char *extra_mounts)
 	if (!collate_premounted_subsystems())
 		return -1;
 
-	build_all_controllers();
+	build_all_controllers(skip_mounts);
 
 	build_controller_mntlist();
 	print_debug_controller_info();
@@ -759,6 +825,116 @@ int collect_subsystems(char *extra_mounts)
 	return 0;
 }
 
+#define NEWROOT "/run/cgmanager/root"
+
+static int do_pivot(void) {
+	int oldroot = -1, newroot = -1;
+
+	oldroot = open("/", O_DIRECTORY | O_RDONLY);
+	if (oldroot < 0) {
+		nih_fatal("%s: Error opening old-/ for fchdir", __func__);
+		return -1;
+	}
+	newroot = open(NEWROOT, O_DIRECTORY | O_RDONLY);
+	if (newroot < 0) {
+		nih_fatal("%s: Error opening new-/ for fchdir", __func__);
+		goto fail;
+	}
+
+	/* change into new root fs */
+	if (fchdir(newroot)) {
+		nih_fatal("%s: can't chdir to new rootfs '%s'", __func__, NEWROOT);
+		goto fail;
+	}
+
+	/* pivot_root into our new root fs */
+	if (pivot_root(".", ".")) {
+		nih_fatal("%s: pivot_root syscall failed: %s",
+				__func__, strerror(errno));
+		goto fail;
+	}
+
+	/*
+	 * at this point the old-root is mounted on top of our new-root
+	 * To unmounted it we must not be chdir'd into it, so escape back
+	 * to old-root
+	 */
+	if (fchdir(oldroot) < 0) {
+		nih_fatal("%s: Error entering oldroot", __func__);
+		goto fail;
+	}
+	if (umount2(".", MNT_DETACH) < 0) {
+		nih_fatal("%s: Error detaching old root", __func__);
+		goto fail;
+	}
+
+	if (fchdir(newroot) < 0) {
+		nih_fatal("%s: Error re-entering newroot", __func__);
+		goto fail;
+	}
+
+	close(oldroot);
+	close(newroot);
+
+	return 0;
+
+fail:
+	if (oldroot != -1)
+		close(oldroot);
+	if (newroot != -1)
+		close(newroot);
+	return -1;
+}
+
+static int pivot_into_new_root(void) {
+	int i, ret;
+	char *createdirs[] = {NEWROOT "/proc", NEWROOT "/run",
+		NEWROOT "/run/cgmanager", NEWROOT "/run/cgmanager/fs", NULL};
+	char path[100];
+
+	/* Mount tmpfs for new root */
+	if (mkdir(NEWROOT, 0755) < 0 && errno != EEXIST) {
+		nih_fatal("%s: Failed to create directory for new root\n", __func__);
+		return -1;
+	}
+	ret = mount("root", NEWROOT, "tmpfs", 0, "size=10000,mode=0755");
+	if (ret < 0) {
+		nih_fatal("%s: Failed to mount tmpfs for root", __func__);
+		return -1;
+	}
+
+	/* create /proc and /run/cgmanager/fs, and move-mount those */
+	for (i = 0; createdirs[i]; i++) {
+		if (mkdir(createdirs[i], 0755) < 0) {
+			nih_fatal("%s: failed to created %s\n", __func__, createdirs[i]);
+			return -1;
+		}
+	}
+
+	ret = snprintf(path, 100, NEWROOT "/proc");
+	if (ret < 0 || ret > 100)
+		return -1;
+	if (mount("/proc", path, NULL, MS_REC|MS_MOVE, 0) < 0) {
+		nih_fatal("%s: failed to move /proc into new root: %s",
+			__func__, strerror(errno));
+		return -1;
+	}
+	ret = snprintf(path, 100, NEWROOT "/run/cgmanager/fs");
+	if (ret < 0 || ret > 100)
+		return -1;
+	if (mount("/run/cgmanager/fs", path, NULL, MS_REC|MS_MOVE, 0) < 0) {
+		nih_fatal("%s: failed to move /run/cgmanager/fs into new root: %s",
+			__func__, strerror(errno));
+		return -1;
+	}
+
+	/* Pivot into new root */
+	if (do_pivot() < 0)
+		return -1;
+
+	return 0;
+}
+
 /**
  * Mount the cgroup filesystems and record the information.
  * This should take configuration data from /etc.  For now,
@@ -780,8 +956,20 @@ int setup_cgroup_mounts(void)
 		return 0;
 	}
 
-	if (mount(NULL, "/", NULL, MS_REC|MS_PRIVATE, 0) < 0)
-		nih_warn("Failed to re-mount / non-shared");
+	if (mount(NULL, "/", NULL, MS_REC|MS_PRIVATE, 0) < 0) {
+		nih_warn("Failed to re-mount / private");
+		return -1;
+	}
+
+	/*
+	 * Mount a tmpfs on top of /root in case / is still ro when we
+	 * started, so we can write the org_freedesktop_general.lock
+	 */
+	if (dir_exists("/root")) {
+		if (mount("root", "/root", "tmpfs", 0, "size=10000") < 0) {
+			nih_warn("Failed to mount a writeable tmpfs onto /root");
+		}
+	}
 
 	for (i=0; i<num_controllers; i++) {
 		if (!do_mount_subsys(i)) {
@@ -790,6 +978,12 @@ int setup_cgroup_mounts(void)
 		}
 	}
 
+	/* Now pivot into a new root */
+	if (pivot_into_new_root() < 0) {
+		nih_fatal("Failed pivoting into new root");
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -1186,6 +1380,8 @@ char *file_read_string(void *parent, const char *path)
 		memset(string+sz-1024, 0, 1024);
 		ret = read(fd, string+sz-1024, 1024);
 		if (ret < 0) {
+			nih_error("failure reading path %s: %s\n",
+				path, strerror(errno));
 			nih_free(string);
 			string = NULL;
 			goto out;
@@ -1389,10 +1585,12 @@ bool chown_cgroup_path(const char *path, uid_t uid, gid_t gid, bool all_children
 		nih_local char *fpath = NULL;
 		fpath = NIH_MUST( nih_sprintf(NULL, "%s/cgroup.procs", path) );
 		if (chown(fpath, uid, gid) < 0)
-			nih_error("Failed to chown procs file %s", fpath);
+			nih_error("Failed to chown procs file %s: %s", fpath,
+				strerror(errno));
 		sprintf(fpath+len, "/tasks");
 		if (chown(fpath, uid, gid) < 0)
-			nih_error("Failed to chown tasks file %s", fpath);
+			nih_error("Failed to chown tasks file %s: %s", fpath,
+				strerror(errno));
 	}
 
 out:
@@ -1452,7 +1650,8 @@ bool set_value_trusted(const char *path, const char *value)
 	if (*value && value[len-1] != '\n')
 		fprintf(f, "\n");
 	if (fclose(f) != 0) {
-		nih_error("Error closing %s", path);
+		nih_error("Error closing %s: %s", path,
+			  strerror(errno));
 		return false;
 	}
 	return true;
@@ -1547,6 +1746,8 @@ bool move_self_to_root(void)
 
 		if (!all_mounts[i].path)
 			continue;
+		if (all_mounts[i].skip)
+			continue;
 		path = NIH_MUST( nih_sprintf(NULL, "%s/tasks", all_mounts[i].path) );
 		if ((f = fopen(path, "w")) == NULL)
 			return false;
@@ -1698,3 +1899,34 @@ bool was_premounted(const char *controller)
 		return false;
 	return all_mounts[i].premounted;
 }
+
+/*
+ * Check that (absolute) @path is under @pid's cgroup for @contr
+ */
+bool path_is_under_taskcg(pid_t pid, const char *contr,const char *path)
+{
+	char pcgpath[MAXPATHLEN];
+	size_t plen;
+
+	// Get p's current cgroup in pcgpath
+	if (!compute_pid_cgroup(pid, contr, "", pcgpath, NULL)) {
+		nih_error("%s: Could not determine the proxy's cgroup for %s",
+				__func__, contr);
+		return false;
+	}
+	plen = strlen(pcgpath);
+	// path must start with pcgpath
+	if (strncmp(pcgpath, path, plen) != 0)
+		return false;
+	// If path is equal to pcgpath then that's ok
+	if (plen == strlen(path))
+		return true;
+	/*
+	 * if path is longer than pcpgath, then it must be a subdirectory
+	 * of pcpgpath. I.e. if pcgpath is /xxx then /xxx/a is ok, /xxx2 is
+	 * not.
+	 */
+	if (path[plen] == '/')
+		return true;
+	return false;
+}
